function Y = Extract_All_Imaging_Biomarkers_StackPSD_3D(foldername, outputfolder, modelfile, winsize,colorimage,idsformat,imageformat,downsample,datasetname,subtypename,tstr,whitenning,numlayers,stepsize,maxthresh,minthresh)
%Function to generate the input samples for RBM..
%Input: foldername - folder containing images
%       winsize - length of image patch/block (assumed to be a square block)
%       colorimage - true for colorimages, false for grayscale
%       idsformat - true if images are in idsformat
%       npatches - number of patches to be extracted from each image
%       (random)
%       downsample - Rescaling the input image (optional)
%       rotation_angle - Angle by which input images have to be rotated
%       before extracting patches.
%Output: Y - A mxn matrix of input data to train the RBM. Here, m is the
%dimension of each input and n is the number of samples.

%Name of the directory to access image files
filenames = dir([foldername '*.' imageformat]);
if not(exist('winsize','var'))
    w = 20;
else
    w = winsize;
end

if ~exist('colorimage','var')
    colorimage = false;
end

if ~exist('downsample','var')
    downsample = 1;
end

Y = 0;
load([modelfile num2str(numlayers) tstr]);

disp(['->files in folder:  ' num2str(length(filenames))]);

status = mkdir([outputfolder num2str(numlayers) '/' subtypename]);
if status == 0
    disp('->Creating folder failed')
    return
else
    disp(['->Creating folder ' outputfolder num2str(numlayers) '/' subtypename ' successfully '])
end

for i = 1:length(filenames)
    
    spsd_fn = [outputfolder num2str(numlayers) '/' subtypename '/' filenames(i).name '_ImagingBiomarkers.txt'];
    
    if exist(spsd_fn,'file')
        disp(['Feature file exists, skipping: ' spsd_fn]);
        continue;
    else
        dlmwrite(spsd_fn,[]); % create an empty file as a placeholder, so other process will bypass it
    end
    
    %Read image
    if idsformat
        I = readIDS([foldername '/' filenames(i).name]);
    elseif strcmp(imageformat,'txt')
        I = readImageTXT([foldername '/' filenames(i).name]);
    else
        I = imread([foldername '/' filenames(i).name]);
    end
    [a1, a2, a3] = size(I);

    I = double(I);
    
    bAbsThresh = -1;
    if( maxthresh > 0 )
        I(find(I>=maxthresh)) = maxthresh;
        I = I - minthresh;
        I(find(I<0)) = 0; 
        I = I./(maxthresh-minthresh);
        bAbsThresh = 1;
    else
        M = I(:);                                      % Create Data
        M = M(find(M>0));
        Ms = sort(M(:),'descend');                          % Sort Descending
        max_median = median(Ms(1:ceil(length(Ms)*0.01)));                % Median of top 1% signal

        max_thresh = max_median;
        if(minthresh >= 0 && minthresh <1)
            min_thresh = max_thresh * minthresh;
        else
            min_thresh = 0;
        end
        I(find(I>=max_thresh)) = max_thresh;
        I = I - min_thresh;
        I(find(I<0)) = 0; 
        I = I./(max_thresh-min_thresh);
        bAbsThresh = 0;
    end
%     I = rescale(I,0,1);

    if exist('downsample','var')
        I = imresize(I,downsample);
    end

    [m1 n1 n2] = size(I);

    patchsize = w;    
    Y1 = [];
    xindex = [];
    yindex = [];
    zindex = [];
    
    for i2 = 1:stepsize:m1-patchsize+1
        for j2 = 1:stepsize:n1-patchsize+1
            for k2 = 1:stepsize:n2-patchsize+1
                I2 = I(i2:i2+patchsize-1,j2:j2+patchsize-1,k2:k2+patchsize-1,:);
                for i1 = 1:w:patchsize-w+1
                    for j1 = 1:w:patchsize-w+1
                        for k1 = 1:w:patchsize-w+1                            
                            xloc = round(i2 + w/2);
                            yloc = round(j2 + w/2);
                            zloc = round(k2 + w/2);
                            if( I(xloc,yloc,zloc) ~= 0 ) % valid image region should have signal != 0
                                xindex = [xindex xloc];
                                yindex = [yindex yloc];
                                zindex = [zindex zloc];
                                I1 = I2(i1:i1+w-1,j1:j1+w-1,k1:k1+w-1,:);
                                Y1 = [Y1 I1(:)];
                            end
                        end
                    end
                end
            end
        end
    end
    
    % Make sure data is zero mean
    if whitenning
        Input = Y1 - repmat(mean_patch,1,size(Y1,2));
        Input = V * Input;
    else
        Input = Y1;
    end
    
    disp([ '->[' num2str(i) '/' num2str(length(filenames)) ']' '   ' filenames(i).name]);
    if(~bAbsThresh)
        disp(['-->Patch size  ' num2str(patchsize) '  Step size  ' num2str(stepsize) '  Absolute threshold  ' num2str(bAbsThresh) '  Max threshold  ' num2str(max_thresh) '  Min threshold  ' num2str(min_thresh)]);
    else
        disp(['-->Patch size  ' num2str(patchsize) '  Step size  ' num2str(stepsize) '  Absolute threshold  ' num2str(bAbsThresh) '  Max threshold  ' num2str(maxthresh) '  Min threshold  ' num2str(minthresh)]);
    end
    disp(['-->Number of sample points  ' num2str(size(Input,2))]);
    
    if( length(Input) == 0 )
        disp('-->No data in mask, ignore...');
        continue;
    end
    
    for layer_idx = 1:numlayers
        Input_intercept = [ones(1,size(Y1,2)) ; Input];
        Input = G_layer{layer_idx,1} * sigmoid(W_layer{layer_idx,1} * Input_intercept);
    end
    mappedX = Input';
    
    disp(['-->Feature file target:   ' spsd_fn]);
    disp(['-->Feature file dimension   number of rows   ' num2str(size(mappedX,1)) '   number of columns   ' num2str(size(mappedX,2))]);
        

    fid = fopen(spsd_fn, 'wt');
    for ii = 1:size(mappedX,1)
        fprintf(fid,'%d \t %d \t %d \t %d \t %d \t %d \t', n1, m1,n2, yindex(ii), xindex(ii), zindex(ii));
        jj = 1;
        while jj <= size(mappedX,2)
            fprintf(fid,'%f \t',mappedX(ii,jj));
            jj = jj + 1;
        end
            fprintf(fid,'\n');
    end
    fclose(fid);

    Y = Y + 1;
end

return
