function [W_layer2 D_layer2 G_layer2 X_layer2] = LayerTrain_v2_server(Input,nbasis,lambda,nepoch,method,optmethod,tauw,initialization)
% Train Predictive Sparse Decomposition
% Input:
%   Input: input features (m x n), feature dim: m, #sample: n
%   nbasis: number of dictionary elements
%   lambda: sparsity, #non-zero elements in the sparse code
%   nepoch: number of iterations
%   method: method for Dictionary Learning, the Dictionary Learning here is
%           used for dictionary initialization
%   optmethod: method for updating W, 
%              optmethod=0 -> ProjW (gradient descent); optmehtod = 1 -> minFunc
%   tauw: learning rate for W
%   initialization: whethter to initialize W as D'
% Output:
%   W_layer2: the encoder for sparse code ( nbasis x m )
%   D_layer2: m x nbasis
%   G_layer2: nbasis x nbasis diag matrix
%   X_layer2: nbasis x n, each column X(:,i) is the sparse code for sample Input(:,i) 

Y_layer2 = Input;
% Y_layer2_intercept = [ones(1,size(Y_layer2,2)) ; Y_layer2];
% nbasis = 200;
% lambda = 50;
% nepoch = 50;
trloss = zeros(nepoch,1);
nepoch
%--------------------------------------------------------------------
DLparam.mode=3;
DLparam.K=nbasis;  % learns a dictionary with 100 elements
DLparam.lambda=lambda;
DLparam.numThreads=7; % number of threads
DLparam.batchsize=2500;%
DLparam.iter=100;
%--------------------------------------------------------------------
if method == 0
    param.L=lambda; % not more than param.L non-zeros coefficients
    param.eps=0.1; % squared norm of the residual should be less than 0.1
    param.numThreads=-1; % number of processors/cores to use; the default choice is -1
                    % and uses all the cores of the machine
elseif method == 1
    param.lambda=0.1; % not more than 20 non-zeros coefficients
    param.numThreads=-1; % number of processors/cores to use; the default choice is -1 % Note: what is the relationship between this paramter and DLparam.numThreads
                         % and uses all the cores of the machine
    param.mode=2;        % penalized formulation
    
end

options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 100;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'none';                    

%--------------------------------------------------------------------
Y = Y_layer2;
D = mexTrainDL(Y,DLparam); % Note: initialization for D ??
D = D * diag(1 ./ sqrt(sum(D.^2)));
[m N] = size(Y);
K = size(D,2);
if initialization == 1
    W = [rand(K,1) D'];% Note: the first column of W is the bias (intercept) vector: b (w0)
else
    W = 0.1*(-0.05 + 0.1*rand(K,m + 1)); 
end
Y_intercept = [ones(1,N) ; Y]; % Note: the first column of Y is the coefficient for the bias (intercept) : b (w0)
% tauw = 0.001;%1e2*0.99*2/norm(Y*Y');

G_old = eye(size(D,2));
W_old = W;
X = [];
for ite = 1:nepoch % start the learning loop
    G = G_old;
    W = W_old;
    SigmaWY = sigmoid(W*Y_intercept);
    
    % update D with minimization objective: || DX - Y ||_F 
    if ~isempty(X)
        X = full(X);% nandita's code, converts a sparse matrix X to full storage organization.  If X is a full matrix, it is left unchanged
        tau_d = 0.99*2/norm(X*X'); % Note: check the FISTA paper out for details.
        for i = 1:50 % Note: ?? why 50 iterations
            D = ProjC2(D,X,Y,tau_d); % Note: documentation missing, update D based on (stochastic) gradient descent
            D = abs(D);
            D = D * diag(1 ./ sqrt(sum(D.^2)));
        end
    end

    % computes X with minimization objective: || DX - Y ||_F + || IX - G SIGMA(WY) ||_F, 
    % which is a combination of Reconstruction and Regression.
    Y_eq = [Y ; G * SigmaWY]; 
    D_eq = [D ; eye(size(D,2))];
    D_eq = D_eq*diag(1./sqrt(sum(D_eq.*D_eq))); % column-wise normalization s.t., each column of D_eq has 2-norm = 1, which is required by SPAMS?
    if method == 0 
        X = mexOMP(Y_eq,D_eq,param); % This is a fast implementation of the Orthogonal Matching Pursuit algorithm, provided by SPAMS
    elseif method == 1
        X = mexLasso(Y_eq,D_eq,param); % This is a fast implementation of the LARS algorithm, provided by SPAMS
    end
    % Note: row-wise normalization. Since we normalize D_eq before solving 
    % X, the solution is for || D_eq_normalized*X - Y||_F
    % rather than || D_eq*X - Y||_F, so we need to normalize (rescale) the solution
    % (X) back to what we desired
    X = diag(1 ./ sqrt(sum([D ; eye(size(D,2))].^2))) * X; 
 
    % compute G, based on X = G SIGMA(WY), basically, it is the least
    % square solution of the over-determined system
    SigmaWY = sigmoid(W*Y_intercept);
    temp = diag(1 ./ sum(SigmaWY .^ 2,2));
    G = diag((temp * SigmaWY)*X');
    G = diag(G);
    
    err_old = norm(G*sigmoid(W*Y_intercept)-X,'fro')/norm(X,'fro');%norm(G*sigmoid(W*Y_intercept)-X,'fro');
    W_old = W;

    % Update W
    if optmethod == 0
        for i = 1:50
            W = ProjW(W,X,Y_intercept,tauw,G);% Note: documentation missing, update W based on (stochastic) gradient descent
        end
        W_new = W;
    elseif optmethod == 1
        theta = W_old(:); % convert matrix to vector by concatenating in column-wise manner; this is required by minFunc                    
        [opttheta, cost] = minFunc( @(p) ObjectFun_MatrixForm(p,X,Y_intercept,G),theta,options); % p wil be replaced by theta
        W_new = reshape(opttheta,size(X,1),size(Y_intercept,1));
    end
    
    disp(['old error: ', num2str(norm(G*sigmoid(W_old*Y_intercept)-X,'fro')), '   old ratio: ' num2str(norm(G*sigmoid(W_old*Y_intercept)-X,'fro')/norm(X,'fro'))]);
    disp(['cur error: ', num2str(norm(G*sigmoid(W_new*Y_intercept)-X,'fro')), '   cur ratio: ' num2str(norm(G*sigmoid(W_new*Y_intercept)-X,'fro')/norm(X,'fro'))]);
    disp(['old - cur =  ', num2str(norm(G*sigmoid(W_old*Y_intercept)-X,'fro')/norm(X,'fro') - norm(G*sigmoid(W_new*Y_intercept)-X,'fro')/norm(X,'fro'))]);
    
    err_cur = norm(G*sigmoid(W_new*Y_intercept)-X,'fro')/norm(X,'fro');%norm(G*sigmoid(W_new*Y_intercept)-X,'fro');
    if (err_cur <= err_old)
        W_old = W_new;
        err_old = err_cur;
        disp(['decreasing'])
    else
        disp(['ascending'])
        tauw = 0.1*tauw;
    end            
        
    trloss(ite) = norm(Y-D(1:size(Y,1),:)*X,'fro')^2 + norm(G*sigmoid(W_old*Y_intercept)-X,'fro')^2;
    disp(['Iteration   ',num2str(ite),'   Object error1 is: ',num2str(trloss(ite)),'   Approximation error ratio is: ',num2str(norm(Y - D * G*sigmoid(W_old*Y_intercept),'fro') / norm(Y,'fro'))]);
    G_old = G;
end
%--------------------------------------------------------------------

W_layer2 = W_old;
G_layer2 = G;
D_layer2 = D;
X_layer2 = X;
